import threading
import customtkinter as ctk
import tkinter as tk
import os
import struct
import json
import time
from tkinter import filedialog


class App(threading.Thread):

    def __init__(self):
        threading.Thread.__init__(self)
        self.prgbar = None
        self.keyfOn = None
        self.objOn = None
        self.file_input = None
        self.animOn = None
        ctk.set_appearance_mode("dark")
        self.root = None
        self.start()

    def callback(self):
        self.root.quit()

    def run(self):
        self.root = ctk.CTk()
        self.root.protocol("WM_DELETE_WINDOW", self.callback)
        self.root.title("Bin Converter")
        self.root.geometry("1032x580")
        self.root.resizable(False, False)

        border = ctk.CTkFrame(master=self.root)  # frame = clean
        border.pack(pady=10, padx=10, fill="both", expand=True)

        frame = ctk.CTkFrame(master=border)  # frame = clean
        frame.pack(pady=10, padx=10)

        frame2 = ctk.CTkFrame(master=border)  # frame = clean
        frame2.pack(pady=10, padx=10)

        title = ctk.CTkLabel(master=frame, text="Input File: ", font=("Nunito", 14))
        title.grid(row=0, column=0, padx=10, pady=20)

        self.file_input = ctk.CTkEntry(master=frame, width=140, placeholder_text="Input File", font=("Nunito", 14))
        self.file_input.grid(row=0, column=1, padx=10, pady=20)

        button = ctk.CTkButton(master=frame, text="Browse", width=80, command=input_file, fg_color="#c51b1b",
                               hover_color="#7f0000", font=("Nunito", 14))
        button.grid(row=0, column=2, padx=10, pady=20)

        do = ctk.CTkButton(master=frame2, anchor="center", text="Convert", command=start, fg_color="#c51b1b",
                           hover_color="#7f0000", font=("Nunito", 20))
        do.pack(pady=20, padx=20)

        self.prgbar = ctk.CTkProgressBar(master=border, orientation="horizontal", mode="indeterminate")
        self.prgbar.pack(padx=20, pady=20)

        self.animOn = ctk.CTkLabel(master=border, text="Animation", fg_color="gray17", font=("Nunito", 14), width=50,
                                   height=28)
        self.animOn.pack(padx=20, pady=20)

        self.objOn = ctk.CTkLabel(master=border, text="Object", fg_color="gray17", font=("Nunito", 14), width=50,
                                  height=28)
        self.objOn.pack(padx=20, pady=20)

        self.keyfOn = ctk.CTkLabel(master=border, text="Keyframe", fg_color="gray17", font=("Nunito", 14), width=50,
                                   height=28)
        self.keyfOn.pack(padx=20, pady=20)

        self.root.mainloop()

    # Add a method to get file_input value from outside the class
    def get_file_input(self):
        return self.file_input.get()

    def get_animon(self):
        return self.animOn.get()

    def text_animon(self, config):
        return self.animOn.configure(text=config)

    def get_objon(self):
        return self.objOn.get()

    def text_objon(self, config):
        return self.objOn.configure(text=config)

    def get_keyfon(self):
        return self.keyfOn.get()

    def text_keyfon(self, config):
        return self.keyfOn.configure(text=config)

    def set_file_input_value(self, value):
        self.file_input.delete(0, tk.END)
        self.file_input.insert(0, value)

    def prgbar_start(self):
        self.prgbar.start()

    def prgbar_stop(self):
        self.prgbar.stop()


app = App()


def input_file():
    global filepath
    filepath = filedialog.askopenfilename(filetypes=[("bin or json", ["*.bin", "*.json"])])
    filepath = os.path.normpath(filepath)
    app.set_file_input_value(filepath)
    return filepath


def calc_space_bytes(input_str):
    pattern = [4, 3, 2, 1]
    index = input_str % len(pattern)
    return pattern[index]


def perform_file_operations():
    try:
        app.prgbar_start()
        file_ext = os.path.splitext(filepath)[1]
    except NameError as e:
        print("NameError", e)
        app.prgbar_stop()

    if file_ext == ".bin":
        unpack_file(filepath)
    elif file_ext == ".json":
        pack_file(filepath)
    else:
        print("Error: Input file is not a json or a bin")


def start():
    global keyframe_total
    threading.Thread(target=perform_file_operations).start()


def unpack_file(file):
    file_name = os.path.splitext(file)[0]
    data = {}
    with open(file, 'rb') as file:
        # Read the number of xmls and add it to data
        num_xmls = struct.unpack('B', file.read(1))[0]
        data["xmls"] = [
            {
                "num_xmls": num_xmls
            }
        ]
        file.seek(3, 1)  # Skip to name_len

        # Add xml data
        for i in range(1, num_xmls + 1):
            name_len = struct.unpack('B', file.read(1))[0] - 1
            file.seek(3, 1)  # Skip to name
            name_bytes = file.read(name_len)
            try:
                name = name_bytes.decode('ascii').rstrip('\x00')
            except UnicodeDecodeError as e:
                print("UnicodeDecodeError:", e)
                break
            # I swear I won't mess up
            file.seek(calc_space_bytes(name_len), 1)
            xml_id = struct.unpack('B', file.read(1))[0]  # THIS MIGHT BE XML ID
            file.seek(7, 1)  # It's always 7 don't worry
            data["xmls"][0]["xml" + str(i)] = [
                {
                    "name": name,
                    "name_len": name_len,
                    "xml_id": xml_id
                }
            ]

        num_anims = struct.unpack('B', file.read(1))[0]  # Read number of animations

        file.seek(3, 1)
        data["animations"] = [
            {
                "num_anims": num_anims
            }
        ]

        # for ii in range(1, num_anims+1):
        for ii in range(1, num_anims + 1):

            name_len = struct.unpack('B', file.read(1))[0] - 1

            # Skip the empty spacer bytes before the Anim name
            file.seek(3, 1)  # Move 3 bytes forward from the current position

            # Read the bytes for the Anim name
            name_bytes = file.read(name_len)

            # Attempt to decode the name from bytes to ASCII
            try:
                name = name_bytes.decode('ascii').rstrip('\x00')
            except UnicodeDecodeError as e:
                print("UnicodeDecodeError:", e)
                break

            file.seek(calc_space_bytes(name_len), 1)

            width = struct.unpack('H', file.read(2))[0]
            height = struct.unpack('H', file.read(2))[0]

            loop_offset = struct.unpack('f', file.read(4))[0]

            centered = struct.unpack('B', file.read(1))[0]

            file.seek(3, 1)

            objects_num = struct.unpack('B', file.read(1))[0]

            file.seek(3, 1)

            data["animations"][0]["anim" + str(ii)] = [
                {
                    "name": name,
                    "name_len": name_len,
                    "width": width,
                    "height": height,
                    "loop_offset": loop_offset,
                    "centered": centered,
                    "objects": [
                        {
                            "num_objects": objects_num
                        }
                    ]
                }
            ]

            if ii == num_anims:
                app.text_animon("Done!")
            else:
                app.text_animon(name)

            # for iii in range(1, 2):
            for iii in range(1, objects_num + 1):
                name_len = struct.unpack('b', file.read(1))[0] - 1
                file.seek(3, 1)
                name_bytes = file.read(name_len)

                # Attempt to decode the name from bytes to ASCII
                try:
                    name = name_bytes.decode('ascii').rstrip('\x00')
                except UnicodeDecodeError as e:
                    print("UnicodeDecodeError:", e)
                    break

                file.seek(calc_space_bytes(name_len), 1)

                object_type = struct.unpack('B', file.read(1))[0]

                file.seek(3, 1)

                blend_type = struct.unpack('B', file.read(1))[0]

                file.seek(3, 1)

                connected_id = struct.unpack('B', file.read(1))[0]
                ref_id = struct.unpack('B', file.read(1))[0]
                object_id = struct.unpack('B', file.read(1))[0]

                file.seek(1, 1)

                xml_id = struct.unpack('B', file.read(1))[0]
                value = struct.unpack('B', file.read(1))[0]

                red = struct.unpack('B', file.read(1))[0]
                file.seek(1, 1)
                green = struct.unpack('B', file.read(1))[0]
                file.seek(1, 1)
                blue = struct.unpack('B', file.read(1))[0]
                file.seek(1, 1)
                _3p0 = struct.unpack('B', file.read(1))[0]

                file.seek(7, 1)

                keyframe_num = struct.unpack('I', file.read(4))[0]

                data["animations"][0]["anim" + str(ii)][0]["objects"][0]["object" + str(iii)] = [
                    {
                        "name": name,
                        "name_len": name_len,
                        "object_type": object_type,
                        "blend_type": blend_type,
                        "connected_id": connected_id,
                        "ref_id": ref_id,
                        "object_id": object_id,
                        "xml_id": xml_id,
                        "value": value,
                        "recolours": [
                            {
                                "red": red,
                                "green": green,
                                "blue": blue,
                            }
                        ],
                        "3p0": _3p0,
                        "keyframes": [
                            {
                                "num_keyframes": keyframe_num
                            }
                        ],
                    }
                ]

                if iii == objects_num:
                    app.text_objon("Done!")
                else:
                    app.text_objon(name)

                mask_x = 0
                mask_y = 0
                mask_w = 0
                mask_h = 0
                anchor_x = 0
                anchor_y = 0
                position_x = 0
                position_y = 0
                scale_x = 0
                scale_y = 0
                rotation = 0
                opacity = 0
                name_len = 1
                name = "0"

                for iiii in range(1, keyframe_num + 1):
                    timecode = struct.unpack('f', file.read(4))

                    mask_update = struct.unpack('b', file.read(1))
                    file.seek(3, 1)
                    mask_x = struct.unpack('f', file.read(4))
                    mask_y = struct.unpack('f', file.read(4))
                    mask_w = struct.unpack('f', file.read(4))
                    mask_h = struct.unpack('f', file.read(4))

                    anchor_update = struct.unpack('b', file.read(1))
                    file.seek(3, 1)
                    anchor_x = struct.unpack('f', file.read(4))
                    anchor_y = struct.unpack('f', file.read(4))

                    position_update = struct.unpack('b', file.read(1))
                    file.seek(3, 1)
                    position_x = struct.unpack('f', file.read(4))
                    position_y = struct.unpack('f', file.read(4))

                    scale_update = struct.unpack('b', file.read(1))
                    file.seek(3, 1)
                    scale_x = struct.unpack('f', file.read(4))
                    scale_y = struct.unpack('f', file.read(4))

                    rotation_update = struct.unpack('b', file.read(1))
                    file.seek(3, 1)
                    rotation = struct.unpack('f', file.read(4))

                    opacity_update = struct.unpack('b', file.read(1))
                    file.seek(3, 1)
                    opacity = struct.unpack('f', file.read(4))

                    sprite_update = struct.unpack('b', file.read(1))
                    if sprite_update == (-1,):
                        file.seek(11, 1)
                    else:
                        file.seek(3, 1)
                        name_len = struct.unpack('B', file.read(1))[0] - 1
                        file.seek(3, 1)
                        name_bytes = file.read(name_len)
                        space_bytes = calc_space_bytes(name_len)
                        file.seek(space_bytes, 1)
                        try:
                            name = name_bytes.decode('ascii').rstrip('\x00')
                        except UnicodeDecodeError as e:
                            print("UnicodeDecodeError:", e)
                            break

                    data["animations"][0]["anim" + str(ii)][0]["objects"][0]["object" + str(iii)][0]["keyframes"][0][
                        "keyframe" + str(iiii)] = [
                        {
                            "timecode": timecode,
                            "mask": [
                                {
                                    "update": mask_update,
                                    "x": mask_x,
                                    "y": mask_y,
                                    "width": mask_w,
                                    "height": mask_h
                                }
                            ],
                            "anchor": [
                                {
                                    "update": anchor_update,
                                    "x": anchor_x,
                                    "y": anchor_y
                                }
                            ],
                            "position": [
                                {
                                    "update": position_update,
                                    "x": position_x,
                                    "y": position_y
                                }
                            ],
                            "scale": [
                                {
                                    "update": scale_update,
                                    "x": scale_x,
                                    "y": scale_y
                                }
                            ],
                            "rotation": [
                                {
                                    "update": rotation_update,
                                    "rotation": rotation
                                }
                            ],
                            "opacity": [
                                {
                                    "update": opacity_update,
                                    "opacity": opacity
                                }
                            ],
                            "sprite": [
                                {
                                    "update": sprite_update,
                                    "name": name,
                                    "name_len": name_len,
                                }
                            ]
                        }
                    ]

                    app.text_keyfon(name)

    # Write the data to the json
    with open(file_name + '.json', 'w') as json_file:
        json.dump(data, json_file, indent=2)

    print("Unpack Success")
    app.text_keyfon("Done!")
    app.prgbar_stop()


def pack_file(file):
    start_time = time.time()
    # Load JSON data from the file
    with open(file, 'r') as json_file:
        data = json.load(json_file)

    file_name = os.path.splitext(file)[0]
    with open(file_name + '.bin', 'wb') as file:
        num_xmls = data["xmls"][0]["num_xmls"]
        file.write(struct.pack('B', num_xmls))
        file.seek(3, 1)
        for i in range(1, num_xmls + 1):
            name_len = data["xmls"][0]["xml" + str(i)][0]["name_len"]
            name = data["xmls"][0]["xml" + str(i)][0]["name"]
            xml_id = data["xmls"][0]["xml" + str(i)][0]["xml_id"]
            file.write(struct.pack('B', (name_len + 1)))
            file.seek(3, 1)
            file.write(name.encode("ascii"))
            file.seek(calc_space_bytes(name_len), 1)
            file.write(struct.pack('B', xml_id))
            file.seek(7, 1)

        num_anims = data["animations"][0]["num_anims"]
        file.write(struct.pack('B', num_anims))
        file.seek(3, 1)

        for ii in range(1, num_anims + 1):
            name_len = data["animations"][0]["anim" + str(ii)][0]["name_len"]
            name = data["animations"][0]["anim" + str(ii)][0]["name"]
            width = data["animations"][0]["anim" + str(ii)][0]["width"]
            height = data["animations"][0]["anim" + str(ii)][0]["height"]
            loop_offset = data["animations"][0]["anim" + str(ii)][0]["loop_offset"]
            centered = data["animations"][0]["anim" + str(ii)][0]["centered"]
            file.write(struct.pack('B', name_len + 1))
            file.seek(3, 1)
            file.write(name.encode("ascii"))
            file.seek(calc_space_bytes(name_len), 1)
            file.write(struct.pack('H', width))
            file.write(struct.pack('H', height))
            file.write(struct.pack('f', loop_offset))
            file.write(struct.pack('B', centered))
            file.seek(3, 1)

            num_objects = data["animations"][0]["anim" + str(ii)][0]["objects"][0]["num_objects"]
            file.write(struct.pack('B', num_objects))
            file.seek(3, 1)

            for iii in range(1, num_objects + 1):
                name_len = data["animations"][0]["anim" + str(ii)][0]["objects"][0]["object" + str(iii)][0]["name_len"]
                name = data["animations"][0]["anim" + str(ii)][0]["objects"][0]["object" + str(iii)][0]["name"]
                object_type = data["animations"][0]["anim" + str(ii)][0]["objects"][0]["object" + str(iii)][0][
                    "object_type"]
                blend_type = data["animations"][0]["anim" + str(ii)][0]["objects"][0]["object" + str(iii)][0][
                    "blend_type"]
                connected_id = data["animations"][0]["anim" + str(ii)][0]["objects"][0]["object" + str(iii)][0][
                    "connected_id"]
                ref_id = data["animations"][0]["anim" + str(ii)][0]["objects"][0]["object" + str(iii)][0]["ref_id"]
                object_id = data["animations"][0]["anim" + str(ii)][0]["objects"][0]["object" + str(iii)][0][
                    "object_id"]
                xml_id = data["animations"][0]["anim" + str(ii)][0]["objects"][0]["object" + str(iii)][0]["xml_id"]
                value = data["animations"][0]["anim" + str(ii)][0]["objects"][0]["object" + str(iii)][0]["value"]
                red = data["animations"][0]["anim" + str(ii)][0]["objects"][0]["object" + str(iii)][0]["recolours"][0][
                    "red"]
                green = \
                    data["animations"][0]["anim" + str(ii)][0]["objects"][0]["object" + str(iii)][0]["recolours"][0][
                        "green"]
                blue = data["animations"][0]["anim" + str(ii)][0]["objects"][0]["object" + str(iii)][0]["recolours"][0][
                    "blue"]
                _3p0 = data["animations"][0]["anim" + str(ii)][0]["objects"][0]["object" + str(iii)][0]["3p0"]

                print(name)

                file.write(struct.pack('B', name_len + 1))
                file.seek(3, 1)
                file.write(name.encode("ascii"))
                file.seek(calc_space_bytes(name_len), 1)
                file.write(struct.pack('B', object_type))
                file.seek(3, 1)
                file.write(struct.pack('B', blend_type))
                file.seek(3, 1)
                file.write(struct.pack('B', connected_id))
                file.write(struct.pack('B', ref_id))
                file.write(struct.pack('B', object_id))
                file.seek(1, 1)
                file.write(struct.pack('B', xml_id))
                file.write(struct.pack('B', value))
                file.write(struct.pack('B', red))
                file.seek(1, 1)
                file.write(struct.pack('B', green))
                file.seek(1, 1)
                file.write(struct.pack('B', blue))
                file.seek(1, 1)
                file.write(struct.pack('B', _3p0))
                file.seek(7, 1)

                num_keyframes = \
                    data["animations"][0]["anim" + str(ii)][0]["objects"][0]["object" + str(iii)][0]["keyframes"][0][
                        "num_keyframes"]
                file.write(struct.pack('B', num_keyframes))
                file.seek(3, 1)

                mask_x = 0
                mask_y = 0
                mask_w = 0
                mask_h = 0
                anchor_x = 0
                anchor_y = 0
                position_x = 0
                position_y = 0
                scale_x = 0
                scale_y = 0
                rotation = 0
                opacity = 0
                name_len = 1
                name = "0"

                for iiii in range(1, num_keyframes + 1):
                    keyframe = \
                        data["animations"][0]["anim" + str(ii)][0]["objects"][0]["object" + str(iii)][0]["keyframes"][
                            0][
                            "keyframe" + str(iiii)]
                    file.write(struct.pack('f', keyframe[0]["timecode"][0]))

                    # MASK BULLSHIT
                    mask_update = keyframe[0]["mask"][0]["update"][0]
                    mask_x = keyframe[0]["mask"][0]["x"][0]
                    mask_y = keyframe[0]["mask"][0]["y"][0]
                    mask_w = keyframe[0]["mask"][0]["width"][0]
                    mask_h = keyframe[0]["mask"][0]["height"][0]
                    file.write(struct.pack('b', mask_update))
                    file.seek(3, 1)
                    if mask_update == -1:
                        file.write(struct.pack('f', 0))
                        file.write(struct.pack('f', 0))
                        file.write(struct.pack('f', 0))
                        file.write(struct.pack('f', 0))
                    else:
                        file.write(struct.pack('f', mask_x))
                        file.write(struct.pack('f', mask_y))
                        file.write(struct.pack('f', mask_w))
                        file.write(struct.pack('f', mask_h))

                    # Anchor
                    anchor_update = keyframe[0]["anchor"][0]["update"][0]
                    anchor_x = keyframe[0]["anchor"][0]["x"][0]
                    anchor_y = keyframe[0]["anchor"][0]["y"][0]
                    file.write(struct.pack('b', anchor_update))
                    file.seek(3, 1)
                    if anchor_update == -1:
                        file.write(struct.pack('f', 0))
                        file.write(struct.pack('f', 0))
                    else:
                        file.write(struct.pack('f', anchor_x))
                        file.write(struct.pack('f', anchor_y))

                    # Position
                    position_update = keyframe[0]["position"][0]["update"][0]
                    position_x = keyframe[0]["position"][0]["x"][0]
                    position_y = keyframe[0]["position"][0]["y"][0]
                    file.write(struct.pack('b', position_update))
                    file.seek(3, 1)
                    if position_update == -1:
                        file.write(struct.pack('f', 0))
                        file.write(struct.pack('f', 0))
                    else:
                        file.write(struct.pack('f', position_x))
                        file.write(struct.pack('f', position_y))

                    # Scale
                    scale_update = keyframe[0]["scale"][0]["update"][0]
                    scale_x = keyframe[0]["scale"][0]["x"][0]
                    scale_y = keyframe[0]["scale"][0]["y"][0]
                    file.write(struct.pack('b', scale_update))
                    file.seek(3, 1)
                    if scale_update == -1:
                        file.write(struct.pack('f', 0))
                        file.write(struct.pack('f', 0))
                    else:
                        file.write(struct.pack('f', scale_x))
                        file.write(struct.pack('f', scale_y))

                    # rotation
                    rotation_update = keyframe[0]["rotation"][0]["update"][0]
                    rotation = keyframe[0]["rotation"][0]["rotation"][0]
                    file.write(struct.pack('b', rotation_update))
                    file.seek(3, 1)
                    if rotation_update == -1:
                        file.write(struct.pack('f', 0))
                    else:
                        file.write(struct.pack('f', rotation))

                    # opacity
                    opacity_update = keyframe[0]["opacity"][0]["update"][0]
                    opacity = keyframe[0]["opacity"][0]["opacity"][0]
                    file.write(struct.pack('b', opacity_update))
                    file.seek(3, 1)
                    if opacity_update == -1:
                        file.write(struct.pack('f', 0))
                    else:
                        file.write(struct.pack('f', opacity))

                    # sprite
                    sprite_update = keyframe[0]["sprite"][0]["update"][0]
                    name_len = keyframe[0]["sprite"][0]["name_len"]
                    name = keyframe[0]["sprite"][0]["name"]
                    file.write(struct.pack('b', sprite_update))
                    file.seek(3, 1)
                    if sprite_update == -1:
                        file.write(struct.pack('d', 4.94065645841247E-324))
                    else:
                        file.write(struct.pack('B', name_len + 1))
                        file.seek(3, 1)
                        file.write(name.encode("ascii"))
                        file.seek(calc_space_bytes(name_len), 1)

    print("Done!")
    print()
    end_time = time.time()
    print("Elapsed time", end_time - start_time)
